﻿////------------------------------------------------------------------------------------------
//
//    MindFusion Office Floorplan Designer 
//     Copyright (c) 2022, MindFusion LLC - Bulgaria
//
////------------------------------------------------------------------------------------------

using MindFusion.Diagramming;

using System;
using System.Drawing;
using System.Xml;


namespace FloorplanDesigner
{
   /// <summary>
   /// The FloorplanDiagram class is derived from MindFusion.Diagramming.Diagram. The reason for 
   /// doing it this way is that it provides an easy way of persisting the floorplan diagrams to XML 
   /// disk files using the MindFusion built-in serialization/deserialization support. (For an 
   /// alternative technique see the import/export functionality in the PasteCopy sample program.)
   /// 
   /// This class contains seven extra fields, as seen in the FormSettings dialog, to provide 
   /// information specific to how this floorplan is to be handled. This includes things like the 
   /// specification of whether the real-world dimensions are in feet or meters, how to convert them 
   /// to the internal dimensions in pixels, information about the background image, etc.
   /// 
   /// These extra fields are persisted along with the rest of the diagram data by adding additional 
   /// XML elements to the XML document object.
   /// </summary>
   public class FloorplanDiagram : Diagram
   {
      #region Constants

      // Width of a door in pixels, and length of minimum wall segment due to door placement
      private const float CDoorWidth = 50f;
      private const float CMinimumWall = 10f;

      // XML element names for the added fields
      private const string CRealWorldUnits = "RealWorldUnits";
      private const string CScalingFactor = "ScalingFactor";
      private const string CRealWorldSizeX = "RealWorldSizeX";
      private const string CRealWorldSizeY = "RealWorldSizeY";
      private const string CBackgroundImageFile = "BackgroundImageFile";
      private const string CBackgroundScaling = "BackgroundScaling";
      private const string CBackgroundOpacity = "BackgroundOpacity";

      // Default values for scaling factor and background image opacity
      public const int CDefaultScalingForFeet = 16;
      public const int CDefaultScalingForMeters = 50;
      public const int CDefaultBackgroundOpacity = 30;

      // Due to use of "one-garian notation" ...
      private const string CEnumPrefix = "e";

      #endregion Constants


      #region Private variables

      // Additional fields added to the Diagram object - need to be persisted

      // Is this floorplan based on feet or meters?
      private ERealWorldUnits _realWorldUnits = ERealWorldUnits.eMeters;

      // Scaling factor from feet or meters to internal dimensions in pixels
      private int _scalingFactor = CDefaultScalingForMeters;

      // Nominal size of the floorplan in "real world units" (meters or feet), zero = not specified
      private int _realWorldSizeX = 0;
      private int _realWorldSizeY = 0;

      // Full path and filename of the optional background image, plus scaling and opacity to be 
      //  applied if there is a background image
      private string _backgroundImageFile = "";  // "" = no background image
      private int _backgroundScaling = 100;  // Percent, can be > 100
      private int _backgroundOpacity = CDefaultBackgroundOpacity;  // Percent, 0 - 100, 0 = off

      // Following working fields do not need to be persisted

      // Counter to avoid double processing in the override methods
      private int _nestCounter = 0;
      
      // Reference to FormMain
      private readonly FormMain _formMain;

      #endregion Private variables


      #region Public properties

      public ERealWorldUnits RealWorldUnits
      {
         get { return _realWorldUnits; }
         set { _realWorldUnits = value; }
      }


      /// <summary>
      /// Get/set scaling factor. Setting to -1 means default value for current "real world units".
      /// </summary>
      public int ScalingFactor
      {
         get { return _scalingFactor; }

         set
         {
            if (value > 0)
               _scalingFactor = value;
            else
            {
               switch (_realWorldUnits)
               {
                  case ERealWorldUnits.eFeet:    _scalingFactor = CDefaultScalingForFeet;    break;
                  case ERealWorldUnits.eMeters:  _scalingFactor = CDefaultScalingForMeters;  break;
               }
            }
         }
      }

      public int RealWorldSizeX
      {
         get { return _realWorldSizeX; }
         set { _realWorldSizeX = value; }
      }

      public int RealWorldSizeY
      {
         get { return _realWorldSizeY; }
         set { _realWorldSizeY = value; }
      }

      public string BackgroundImageFile
      {
         get { return _backgroundImageFile; }
         set { _backgroundImageFile = value; }
      }

      public int BackgroundScaling
      {
         get { return _backgroundScaling; }
         set { _backgroundScaling = value; }
      }

      public int BackgroundOpacity
      {
         get { return _backgroundOpacity; }
         set { _backgroundOpacity = value; }
      }

      #endregion Public properties


      #region Constructor

      /// <summary>
      /// Constructor.
      /// </summary>
      public FloorplanDiagram(FormMain formMain)
      {
         // Save argument and hook up a couple of Clicked event handler
         _formMain = formMain;
         this.Clicked += Diagram_Clicked;
         this.NodeClicked += Diagram_NodeClicked;

         // Set some non-default properties. But note that changing these will not be effective for 
         //  floorplan files that have previously been saved - fetching them will override these 
         //  values unless code is added to the FormMain.File_Open_Click() method to do otherwise. 
         this.AdjustmentHandlesSize = 6f;
		 this.MeasureUnit = MeasureUnit.Pixel;
         this.Dirty = false;
      }

      #endregion Constructor


      #region Public override methods used to support persisting to/from an XML file

      /// <summary>
      /// This method apparently does not get called, but it is overridden anyway just in case. 
      /// </summary>
      public override void LoadFromXml(System.Xml.XmlDocument xmlDocument)
      {
         _nestCounter++;
         base.LoadFromXml(xmlDocument);
         if (_nestCounter == 1)
            LoadExtraFieldsFromXml(xmlDocument);
         _nestCounter--;
      }


      /// <summary>
      /// Method called by the base object when LoadFromXml(fileName) has been used and the XML 
      /// document object is available.
      /// </summary>
      public override void LoadFromXml(XmlDocument xmlDocument, bool loadDiagramProperties)
      {
         _nestCounter++;
         base.LoadFromXml(xmlDocument, loadDiagramProperties);
         if (_nestCounter == 1)
            LoadExtraFieldsFromXml(xmlDocument);
         _nestCounter--;
      }


      /// <summary>
      /// Method called by the base object when SaveToXml(fileName) has been used and an XML 
      /// document object has been created..
      /// </summary>
      public override void SaveToXml(XmlDocument xmlDocument)
      {
         _nestCounter++;
         base.SaveToXml(xmlDocument);
         if (_nestCounter == 1)
            SaveExtraFieldsToXml(xmlDocument);
         _nestCounter--;
      }


      /// <summary>
      /// This method apparently gets called when the above method calls base.SaveToXml(). That's 
      /// why there is the extra code involving _nestCounter, otherwise the added fields would get 
      /// added twice to the XML document.
      /// </summary>
      public override void SaveToXml(XmlDocument xmlDocument, bool includeUnalteredProperties)
      {
         _nestCounter++;
         base.SaveToXml(xmlDocument, includeUnalteredProperties);
         if (_nestCounter == 1)
            SaveExtraFieldsToXml(xmlDocument);
         _nestCounter--;
      }

      #endregion Public override methods used to support persisting to/from an XML file


      #region Private methods
      
      /// <summary>
      /// Method called when user clicks on the floorplan area where there is no item. This creates 
      /// a floorplan item and places it on the floorplan. But this is not applicable to all items, 
      /// for example walls.
      /// </summary>
      private void Diagram_Clicked(object sender, DiagramEventArgs e)
      {
         // This code only needed because of "pseudo margins" - remove if that is deprecated
         if (e.MousePosition.X < FormMain.CPseudoMarginZero || 
             e.MousePosition.Y < FormMain.CPseudoMarginZero)
            return;

         // Ignore click that was used to end inplace editing mode. This is pretty kludgy, and 
         //  there may well be a better way of doing this.
         if (Environment.TickCount - _formMain._InplaceEditingEnded < 500)
            return;

         // Create a new node object according to checked button on toolbar, check "droppable"
         NodeItemBase nodeItem = _formMain.GetNewNode(e.MousePosition, ENodeRequirement.eDroppable);
         if (nodeItem == null)
            return;  // Exit if error (message has been displayed)

         // Add this node to the diagram unless it is a door object, which is just a fake node
         if (nodeItem is NodeDoor)
            ProcessDoorNode(e.MousePosition);
         else 
            this.Nodes.Add(nodeItem);
      }


      /// <summary>
      /// Internal method of above method to process the user dropping a door node onto the diagram. 
      /// This finds the wall that is nearest the drop point, checks that it is near enough, and 
      /// then calls the method that does the actual processing of placing a door in a wall. 
      /// 
      /// This method is overkill - it would be acceptable to demand that the user actually drop the 
      /// door on the wall instead of just near it. Also, the current tolerance is so small that it 
      /// is difficult to miss the wall but still be close enough for this code to accept it.
      /// </summary>
      private void ProcessDoorNode(PointF dropPosition)
      {
         NodeWallBase closestWall = null;
         float closestGap = float.MaxValue;

         // Loop to check all of the wall nodes
         foreach (DiagramNode diagramNode in this.Nodes)
         {
            // Check that it is a wall node
            NodeWallBase nodeWallBase = diagramNode as NodeWallBase;
            if (nodeWallBase == null)
               continue;

            // Compute distance from drop point to the center of the wall
            float halfWidth = nodeWallBase.WallWidth / 2f;
            float dropDelta;
            if (nodeWallBase.IsVertical)
               dropDelta = Math.Abs(dropPosition.X - (nodeWallBase.Bounds.X + halfWidth));
            else
               dropDelta = Math.Abs(dropPosition.Y - (nodeWallBase.Bounds.Y + halfWidth));

            // Remember this wall if it is the closest
            if (dropDelta < closestGap)
            {
               closestWall = nodeWallBase;
               closestGap = dropDelta;
            }
         }

         // Check there was a closest wall, and it is close enough 
         if (closestWall != null && closestGap <= 3f * NodeItemBase.CNearnessThreshold)
            ProcessDoorNode(dropPosition, closestWall);
         else 
            FormMain.ShowMessageBoxError("Door must be dropped on a wall.");
      }


      /// <summary>
      /// Method called when user clicks on an existing node on the floorplan. This is only 
      /// applicable when the user has selected the "door" button on the toolbar.
      /// </summary>
      void Diagram_NodeClicked(object sender, NodeEventArgs e)
      {
         // Get and check the wall node item
         NodeWallBase nodeWallBase = e.Node as NodeWallBase;
         if (nodeWallBase == null)
            return;  // Just ignore clicks on desks and tables

         // Check that the "door" button is checked on the toolbar, do door processing if so
         NodeItemBase nodeItem = _formMain.GetNewNode(e.MousePosition, ENodeRequirement.eUnknown);
         if (nodeItem != null && nodeItem is NodeDoor)
            ProcessDoorNode(e.MousePosition, nodeWallBase);
      }


      /// <summary>
      /// Method to process the user dropping a door node onto the diagram, either on a wall item or 
      /// very near to one. This creates a "hole" in the wall by replacing the current wall item 
      /// with one or two new shorter ones. Or, for a very short wall, it may simply delete the wall 
      /// node.
      /// </summary>
      private void ProcessDoorNode(PointF dropPosition, NodeWallBase nodeWallBase)
      {
         // Create two, one or zero wall segments to replace the existing wall. The processing for 
         //  vertical and horizontal walls is identical, but needs to take the orientation into 
         //  consideration. To simplify things some of the processing is done in sub-methods on 
         //  numbers that are orientation neutral.
         float halfWidth = nodeWallBase.WallWidth / 2f;
         float newLength1, newLength2;
         if (nodeWallBase.IsVertical)
         {
            NewWallLengths(nodeWallBase.Bounds.Top, nodeWallBase.Bounds.Bottom, dropPosition.Y,
                           halfWidth, out newLength1, out newLength2);
            CreateNewWall(nodeWallBase, nodeWallBase.Bounds.X + halfWidth,
                          nodeWallBase.Bounds.Y + halfWidth, newLength1);
            CreateNewWall(nodeWallBase, nodeWallBase.Bounds.X + halfWidth,
                          nodeWallBase.Bounds.Bottom - halfWidth - newLength2, newLength2);
         }
         else
         {
            NewWallLengths(nodeWallBase.Bounds.Left, nodeWallBase.Bounds.Right, dropPosition.X,
                           halfWidth, out newLength1, out newLength2);
            CreateNewWall(nodeWallBase, nodeWallBase.Bounds.X + halfWidth,
                          nodeWallBase.Bounds.Y + halfWidth, newLength1);
            CreateNewWall(nodeWallBase, nodeWallBase.Bounds.Right - halfWidth - newLength2,
                          nodeWallBase.Bounds.Y + halfWidth, newLength2);
         }

         // Remove the original wall node
         this.Nodes.Remove(nodeWallBase);
      }


      /// <summary>
      /// Internal method of above method to calculate the lengths of the two new wall segments to 
      /// replace the original wall. These calculations are independent of the wall's orientation 
      /// (horizontal or vertical).
      /// </summary>
      private static void NewWallLengths(float wallStart, float wallEnd, float doorCenter, 
                                        float halfWidth, out float newLength1, out float newLength2)
      {
         newLength1 = (doorCenter - CDoorWidth / 2f) - (wallStart + halfWidth);
         newLength2 = (wallEnd - halfWidth) - (doorCenter + CDoorWidth / 2f);
      }


      /// <summary>
      /// Internal method of ProcessDoorNode() to create a new wall segment to replace the existing 
      /// wall that the door was dropped on. This is called twice, but these two calls may result in 
      /// zero or one or two new walls being created.
      /// </summary>
      private void CreateNewWall(NodeWallBase existingWall, float newX, float newY, float newLength)
      {
         if (newLength < CMinimumWall)
            return;

         PointF startPoint = new PointF(newX, newY);
         if (existingWall is NodeWallExternal)
            this.Nodes.Add(new NodeWallExternal(startPoint, (int)newLength, existingWall.IsVertical));
         else 
            this.Nodes.Add(new NodeWallInternal(startPoint, (int)newLength, existingWall.IsVertical));
      }
      

      /// <summary>
      /// Method called by the LoadFromXml() override methods to extract the extra fields this 
      /// derived object adds to the MindFusion Diagram object. 
      /// </summary>
      private void LoadExtraFieldsFromXml(XmlDocument xmlDocument)
      {
         _realWorldUnits = LoadEnumFromXml(xmlDocument, CRealWorldUnits, ERealWorldUnits.eMeters);
         ScalingFactor = LoadIntFromXml(xmlDocument, CScalingFactor, -1);
         _realWorldSizeX = LoadIntFromXml(xmlDocument, CRealWorldSizeX, 0);
         _realWorldSizeY = LoadIntFromXml(xmlDocument, CRealWorldSizeY, 0);
         _backgroundImageFile = LoadStringFromXml(xmlDocument, CBackgroundImageFile);
         _backgroundScaling = LoadIntFromXml(xmlDocument, CBackgroundScaling, 100);
         _backgroundOpacity = 
                      LoadIntFromXml(xmlDocument, CBackgroundOpacity, CDefaultBackgroundOpacity);
      }


      /// <summary>
      /// Method to get an XML element and convert it to an enum value.
      /// See  http://stackoverflow.com/questions/1082532/how-to-tryparse-for-enum-value
      /// </summary>
      private static TEnum LoadEnumFromXml<TEnum>(XmlDocument xmlDocument, string nodeName, 
                                                  TEnum defaultValue) 
      {
         string s = LoadStringFromXml(xmlDocument, nodeName);
         if (s == null)
            return defaultValue;
         if (s.Length >= 2 && char.IsUpper(s[0]))
         {
            s = CEnumPrefix + s;
            if (!Enum.IsDefined(typeof(TEnum), s)) 
               return defaultValue; 
 
            return (TEnum)Enum.Parse(typeof(TEnum), s); 
         }

         InvalidNodeError(nodeName);
         return defaultValue;
      }


      /// <summary>
      /// Method to get an XML element and convert it to an int.
      /// </summary>
      private static int LoadIntFromXml(XmlDocument xmlDocument, string nodeName, int defaultValue) 
      {
         string s = LoadStringFromXml(xmlDocument, nodeName);
         if (s == null)
            return defaultValue;

         int i;
         if (int.TryParse(s, out i))
            return i;

         InvalidNodeError(nodeName);
         return defaultValue;
      }


      /// <summary>
      /// Method to display an "invalid node" error message.
      /// </summary>
      private static void InvalidNodeError(string nodeName)
      {
         FormMain.ShowMessageBoxError("Invalid node in XML file: " + nodeName);
      }


      /// <summary>
      /// Method to get a string value from an XML element.
      /// </summary>
      /// <returns>string value, or null if error</returns>
      private static string LoadStringFromXml(XmlDocument xmlDocument, string nodeName)
      {
         XmlNode xmlNode = xmlDocument.SelectSingleNode("//*/" + nodeName);
         if (xmlNode != null)
            return xmlNode.InnerText;

         FormMain.ShowMessageBoxError("Node missing in XML file: " + nodeName);
         return null;
      }


      /// <summary>
      /// Method called by the SaveToXml() override methods to copy the extra fields this derived 
      /// object adds to the MindFusion Diagram object to the XML document.
      /// </summary>
      private void SaveExtraFieldsToXml(XmlDocument xmlDocument)
      {
         SaveEnumToXml(xmlDocument, CRealWorldUnits, _realWorldUnits);
         SaveStringToXml(xmlDocument, CScalingFactor, _scalingFactor.ToString());
         SaveStringToXml(xmlDocument, CRealWorldSizeX, _realWorldSizeX.ToString());
         SaveStringToXml(xmlDocument, CRealWorldSizeY, _realWorldSizeY.ToString());
         SaveStringToXml(xmlDocument, CBackgroundImageFile, _backgroundImageFile);
         SaveStringToXml(xmlDocument, CBackgroundScaling, _backgroundScaling.ToString());
         SaveStringToXml(xmlDocument, CBackgroundOpacity, _backgroundOpacity.ToString());
      }


      /// <summary>
      /// Method to add an XML element to an XML document containing an enum value.
      /// </summary>
      private static void SaveEnumToXml<TEnum>(XmlDocument xmlDocument, string nodeName, 
                                               TEnum enumValue)
      {
         string s = Enum.GetName(typeof(TEnum), enumValue);
         SaveStringToXml(xmlDocument, nodeName, s.Substring(CEnumPrefix.Length));
      }


      /// <summary>
      /// Method to save a string value as an XML element.
      /// </summary>
      private static void SaveStringToXml(XmlDocument xmlDocument, string nodeName, 
                                          string stringValue)
      {
         XmlNode xmlNode = xmlDocument.CreateElement(nodeName);
         xmlNode.InnerText = stringValue;
         if (xmlDocument.DocumentElement != null)  // Should never happen
            xmlDocument.DocumentElement.AppendChild(xmlNode);
      }

      #endregion Private methods

   }
}
