﻿////------------------------------------------------------------------------------------------
//
//    MindFusion Office Floorplan Designer 
//     Copyright (c) 2022, MindFusion LLC - Bulgaria
//
////------------------------------------------------------------------------------------------

using MindFusion.Diagramming;
using MindFusion.Drawing;

using MindFusionPen = MindFusion.Drawing.Pen;
using MindFusionSolidBrush = MindFusion.Drawing.SolidBrush;

using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Xml;

using SystemPen = System.Drawing.Pen;


namespace FloorplanDesigner
{
	/// <summary>
	/// This is a base class from which NodeWallExternal and NodeWallInternal are derived. 
	/// 
	/// During interactive creation the bounds for the item are inclusive of the entire area bounded 
	/// by the starting point and the current mouse location, where the starting point is one corner 
	/// of the rectangle and the current mouse location is the opposite corner of this invisible 
	/// rectangle. The wall is drawn from the starting point, but not towards the current mouse 
	/// location; instead it is drawn horizontally or vertically along one of the sides of the 
	/// rectangle, namely whichever side is the longest. As the user moves the mouse around the 
	/// rectangle is constantly changing size in both directions, and the side which is longest may 
	/// change, and the wall will then jump from being vertical to horizontal, or vice versa.
	/// 
	/// At the end of interactive creation the bounds are recalculated to correspond to the area that 
	/// the wall actually occupies on the drawing area.
	/// 
	/// One thing to note is that the wall is drawn as being half a width longer at each end, 
	/// compared to the nominal starting and ending points, and that this is reflected by the 
	/// dimensions of the bounds, except during interactive creation. This is done to make the 
	/// corners match up properly when a new wall is drawn at right angles from the end of an 
	/// existing wall, or when the walls are created via programming.
	/// 
	/// However, this is poorly implemented - it needs to be made more object-oriented so the actual 
	/// coordinates are available as properties instead of being recalculated again and again using 
	/// the half width factor.
	/// </summary>
	public abstract class NodeWallBase : NodeItemBase
	{
		#region Constants

		// XML element names for the added fields
		private const string CIsVertical = "IsVertical";
		private const string CWallWidth = "WallWidth";

		#endregion Constants


		#region Private variables

		// These two fields are persisted when saving to a file or copying to the clipboard. (This is 
		//  not really necessary - these two fields could be recalculated from the Bounds values, but 
		//  this was first realized late in the development cycle. Anyway, now they provide an example 
		//  of how to persist custom item data.)

		// Orientation of the wall, horizontal or vertical (in the 2D floorplan sense). This is first 
		//  valid after creation of the wall is complete, and can not be changed after that.
		private bool _isVertical;

		// Width of the wall, external walls are wider than internal walls
		private float _wallWidth;

		// These do not need persisting

		// Switch indicating that the wall node item is being created interactively by the user
		private bool _isCreation = false;

		// The starting point when user is creating the wall node item. This value is not valid after 
		//  user creation is complete, and is never valid for creation by the program.
		private PointF _startPoint;

		#endregion Private variables


		#region Public properties

		public bool IsVertical
		{ get { return _isVertical; } }


		public float WallWidth
		{ get { return _wallWidth; } }

		#endregion Public properties


		#region Constructors

		// Copy constructor, needed for copy to/from clipboard
		protected NodeWallBase(NodeWallBase copyFrom)
			: base(copyFrom)
		{
			_isVertical = copyFrom._isVertical;
			_wallWidth = copyFrom._wallWidth;
		}


		// Constructor used for various purposes
		protected NodeWallBase(Color wallColor, float wallWidth,
							   PointF startPoint, int initialWallLength, bool isVertical)
		{
			// Save start point for interactive or program creation, PointF.Empty otherwise
			_startPoint = startPoint;

			// Customize according to which kind of wall node item (external or internal) 
			_wallWidth = wallWidth;
			base.Pen = new MindFusionPen(wallColor, wallWidth);
			base.Brush = new MindFusionSolidBrush(wallColor);

			// Only three handles will be drawn, not nine
			base.HandlesStyle = HandlesStyle.Custom;

			// Set the item bounds here and now for creation via programming
			if (initialWallLength != 0)
			{
				if (initialWallLength < 0)
					throw new ArgumentException("Wall length must be positive.", "initialWallLength");

				_isVertical = isVertical;
				float halfWidth = wallWidth / 2f;

				if (isVertical)
				{
					base.Bounds = RectangleF.FromLTRB(startPoint.X - halfWidth, startPoint.Y - halfWidth,
								 startPoint.X + halfWidth, startPoint.Y + initialWallLength + halfWidth);
				}
				else
				{
					base.Bounds = RectangleF.FromLTRB(startPoint.X - halfWidth, startPoint.Y - halfWidth,
								 startPoint.X + initialWallLength + halfWidth, startPoint.Y + halfWidth);
				}
			}
		}

		#endregion Constructors


		#region Public and protected override methods

		/// <summary>
		/// Override for the StartCreate() method. This just notes that the wall is being created 
		/// interactively. (Start point has already been saved by the constructor.)
		/// </summary>
		protected override void StartCreate(PointF startPoint)
		{
			base.StartCreate(startPoint);

			_isCreation = true;
		}


		/// <summary>
		/// Override for the Draw() method to draw the wall node item. It is simply drawn as a thick 
		/// line, thickness and color depending on whether internal or external wall.
		/// </summary>
		public override void Draw(IGraphics iGraphics, RenderOptions renderOptions)
		{
			PointF lineStart;
			PointF lineEnd;
			float halfWidth = _wallWidth / 2f;

			if (_isCreation)
			{
				// During interactive creation by the user the bounds describe a rectangle with the 
				//  start point in one corner and the current mouse position in the opposite corner. The 
				//  wall is drawn along one of the sides of the rectangle, and may change dynamically 
				//  between horizontal and vertical orientation.
				if (base.Bounds.Height > base.Bounds.Width)
				{
					// Vertical wall 
					lineStart = new PointF(_startPoint.X, base.Bounds.Top - halfWidth);
					lineEnd = new PointF(_startPoint.X, base.Bounds.Bottom + halfWidth);
				}
				else
				{
					// Horizontal wall
					lineStart = new PointF(base.Bounds.Left - halfWidth, _startPoint.Y);
					lineEnd = new PointF(base.Bounds.Right + halfWidth, _startPoint.Y);
				}
			}
			else
			{
				// After creation the bounds are used in a fairly straight-forward way
				if (_isVertical)
				{
					// Vertical wall 
					lineStart = new PointF(base.Bounds.X + halfWidth, base.Bounds.Top);
					lineEnd = new PointF(base.Bounds.X + halfWidth, base.Bounds.Bottom);
				}
				else
				{
					// Horizontal wall
					lineStart = new PointF(base.Bounds.Left, base.Bounds.Y + halfWidth);
					lineEnd = new PointF(base.Bounds.Right, base.Bounds.Y + halfWidth);
				}
			}

			using (SystemPen systemPen = base.Pen.CreateGdiPen(iGraphics))
			{
				iGraphics.DrawLine(systemPen, lineStart.X, lineStart.Y, lineEnd.X, lineEnd.Y);
			}
		}


		/// <summary>
		/// Override for the CompleteCreate() method. This sets the _isVertical and _isCreation 
		/// fields, as well as setting the bounds in the base object to describe the actual bounds of 
		/// the wall itself, not the enclosing rectangle that has existed up until now.
		/// </summary>
		protected override void CompleteCreate(PointF endPoint)
		{
			// Prevent wall going into "pseudo margins" area
			if (endPoint.X < FormMain.CPseudoMarginZero || endPoint.Y < FormMain.CPseudoMarginZero)
				endPoint = new PointF(Math.Max(endPoint.X, FormMain.CPseudoMarginZero),
									  Math.Max(endPoint.Y, FormMain.CPseudoMarginZero));

			base.CompleteCreate(endPoint);

			float halfWidth = _wallWidth / 2f;

			if (base.Bounds.Height > base.Bounds.Width)
			{
				// Vertical wall 
				_isVertical = true;
				base.Bounds = RectangleF.FromLTRB(_startPoint.X - halfWidth, base.Bounds.Y - halfWidth,
											 _startPoint.X + halfWidth, base.Bounds.Bottom + halfWidth);
			}
			else
			{
				// Horizontal wall
				_isVertical = false;
				base.Bounds = RectangleF.FromLTRB(base.Bounds.X - halfWidth, _startPoint.Y - halfWidth,
											  base.Bounds.Right + halfWidth, _startPoint.Y + halfWidth);
			}

			_isCreation = false;
		}


		/// <summary>
		/// Override for the DrawHandles() method. This draws a single handle at each end of the wall 
		/// and a single handle in the middle.
		/// </summary>
		public override void DrawHandles(
			IGraphics iGraphics,
			HandlesVisualStyle handlesVisualStyle,
			RenderOptions renderOptions)
		{
			List<PointF> handlesList = new List<PointF>(3);

			float centerX = base.Bounds.Left + base.Bounds.Width / 2;
			float centerY = base.Bounds.Top + base.Bounds.Height / 2;

			handlesList.Add(new PointF(centerX, centerY));

			if (_isVertical)
			{
				handlesList.Add(new PointF(centerX, base.Bounds.Top));
				handlesList.Add(new PointF(centerX, base.Bounds.Bottom));
			}
			else
			{
				handlesList.Add(new PointF(base.Bounds.Left, centerY));
				handlesList.Add(new PointF(base.Bounds.Right, centerY));
			}

			DrawHandles(handlesList, iGraphics, handlesVisualStyle, this.Parent.AdjustmentHandlesSize);
		}


		/// <summary>
		/// Override for the HitTestHandle() method. This uses the EHandleId enum internally, but for 
		/// compatibility purposes returns it as an int value.
		/// </summary>
		public override AdjustmentHandle HitTestHandle(PointF mousePoint, float distance)
		{
			EHandleId handleId = EHandleId.eUnknown;

			if (base.Selected)
			{
				float centerX = base.Bounds.Left + base.Bounds.Width / 2;
				float centerY = base.Bounds.Top + base.Bounds.Height / 2;

				if (_isVertical)
				{
					if (TestForNear(mousePoint, centerX, base.Bounds.Top))
						handleId = EHandleId.eTopCenter;
					else if (TestForNear(mousePoint, centerX, centerY))
						handleId = EHandleId.eMoveHandle;
					else if (TestForNear(mousePoint, centerX, base.Bounds.Bottom))
						handleId = EHandleId.eBottomCenter;
				}
				else
				{
					if (TestForNear(mousePoint, base.Bounds.Left, centerY))
						handleId = EHandleId.eLeftCenter;
					else if (TestForNear(mousePoint, centerX, centerY))
						handleId = EHandleId.eMoveHandle;
					else if (TestForNear(mousePoint, base.Bounds.Right, centerY))
						handleId = EHandleId.eRightCenter;
				}
			}

			if (handleId == EHandleId.eUnknown)
				return null;
			return new NodeAdjustmentHandle((NodeHandleType)handleId);
		}


		/// <summary>
		/// Override for the LoadFrom() method, needed to support loading from binary disk files 
		/// (currently not used), and also copying from the clipboard.
		/// </summary>
		protected override void LoadFrom(BinaryReader binaryReader, PersistContext persistContext)
		{
			base.LoadFrom(binaryReader, persistContext);

			_isVertical = binaryReader.ReadBoolean();
			_wallWidth = binaryReader.ReadSingle();
		}


		/// <summary>
		/// Override for the LoadFrom() method, needed to support copying XML disk files.
		/// </summary>
		protected override void LoadFromXml(XmlElement xmlElement, XmlPersistContext persistContext)
		{
			base.LoadFromXml(xmlElement, persistContext);

			_isVertical = persistContext.ReadBool(CIsVertical, xmlElement);
			_wallWidth = persistContext.ReadFloat(CWallWidth, xmlElement);
		}


		/// <summary>
		/// Override for the SaveTo() method, needed to support saving to binary disk files (currently 
		/// not used), and also copying to the clipboard.
		/// </summary>
		protected override void SaveTo(BinaryWriter binaryWriter, PersistContext persistContext)
		{
			base.SaveTo(binaryWriter, persistContext);

			binaryWriter.Write(_isVertical);
			binaryWriter.Write(_wallWidth);
		}


		/// <summary>
		/// Override for the SaveToXml() method, needed to support saving to XML disk files.
		/// </summary>
		protected override void SaveToXml(XmlElement xmlElement, XmlPersistContext persistContext)
		{
			base.SaveToXml(xmlElement, persistContext);

			persistContext.WriteBool(_isVertical, CIsVertical, xmlElement);
			persistContext.WriteFloat(_wallWidth, CWallWidth, xmlElement);
		}

		#endregion Public and protected override methods

	}
}