﻿////------------------------------------------------------------------------------------------
//
//    MindFusion Office Floorplan Designer 
//     Copyright (c) 2022, MindFusion LLC - Bulgaria
//
////------------------------------------------------------------------------------------------

using MindFusion.Diagramming;
using MindFusion.Diagramming.WinForms;

using System;
using System.Drawing;
using System.IO;
using System.Windows.Forms;

using SystemPath = System.IO.Path;


namespace FloorplanDesigner
{
	/// <summary>
	/// FormMain is the main form (surprise, surprise) of the MindFusion Floorplan Designer program. 
	/// 
	/// This program is intended to be a "proof of concept" application, showing that it is possible 
	/// to create an office floorplan designer style of program using the MindFusion Diagramming for 
	/// WinForms product, aka MindFusion FlowChart.Net. But this program is nowhere near being a 
	/// finished or useable application.
	/// 
	/// See the documentation text file for more information.
	/// </summary>
	public partial class FormMain : Form
	{
		#region Constants

		public const string CProgramName = "MindFusion.Diagramming Sample: Floorplan Designer";

		private const string CFileNameExtension = ".mfp";  // MindFusion FloorPlan

		private const float CPseudoMarginSize = 20f;  // See comments on EstablishPseudoMargins()
		public const float CPseudoMarginZero = 0f;  // For symbolic references from dependent code

		private const float CPasteOffset = 10f;  // See comments on _pasteOffset

		#endregion Constants


		#region Private variables

		// Floorplan file name, or null if floorplan file name not specified yet. If not null the file 
		//  name should always include a file name extension and a full path.
		private string _floorplanFilename = null;

		// X and Y offset for pasting items from the clipboard relative to their original position. 
		//  This is done to make it more obvious that the paste operation has worked - if the items 
		//  are pasted on top of existing items you could mistakenly think that nothing has happened.
		private float _pasteOffset = CPasteOffset;

		// Tick count when inplace editing ended
		private int _inplaceEditingEnded = 0;

		#endregion Private variables


		internal int _InplaceEditingEnded
		{ get { return _inplaceEditingEnded; } }


		#region Constructor

		/// <summary>
		/// Constructor.
		/// </summary>
		public FormMain()
		{
			// Initialize the Visual Studio designer-generated stuff, including the MindFusion 
			//  DiagramView and Ruler controls
			// ReSharper disable DoNotCallOverridableMethodsInConstructor
			InitializeComponent();
			this.Text = CProgramName;  // Just in case the constant is changed
			// ReSharper restore DoNotCallOverridableMethodsInConstructor

			// Instantiate an empty FloorplanDiagram object and connect it to the DiagramView object, 
			//  then set up the ruler to show meters (default real-world units, default scaling)
			diagramView.Diagram = new FloorplanDiagram(this);
			ApplySettings();

			// Hook up the custom "behavior" class to the DiagramView control
			diagramView.Behavior = Behavior.Custom;
			diagramView.CustomBehavior = new FloorplanBehavior(this, diagramView);

			// Register the custom node item classes for serialization/deserialization
			Diagram.RegisterItemClass(typeof(NodeDesk), "Floorplan.Desk", 1);
			Diagram.RegisterItemClass(typeof(NodeTable), "Floorplan.Table", 1);
			Diagram.RegisterItemClass(typeof(NodeTextBox), "Floorplan.TextBox", 1);
			Diagram.RegisterItemClass(typeof(NodeWallExternal), "Floorplan.WallExternal", 1);
			Diagram.RegisterItemClass(typeof(NodeWallInternal), "Floorplan.WallInternal", 1);

			// Create 20 pixel "pseudo margins" - see comments on the EstablishPseudoMargins() method
			EstablishPseudoMargins();
            
			if (File.Exists("SamplePlan.mfp"))
				LoadFloorplan("SamplePlan.mfp");

            diagramView.Diagram.BackBrush = new MindFusion.Drawing.SolidBrush(Color.FromArgb(224, 233, 233));

		}

		#endregion Constructor


		#region Public methods

		/// <summary>
		/// Method to instantiate a new floorplan item of the type selected on the ToolStrip. This is 
		/// called by both Behavior.StartDraw() and Diagram_Clicked(), although not all floorplan 
		/// items are suitable for drawing (i.e., call by Behavior.StartDraw()) or being "dropped" 
		/// (i.e., call by Diagram_Clicked()).
		/// </summary>
		/// <param name="startPoint">current location of mouse</param>
		/// <param name="nodeRequirement">does current node have to be drawable or droppable?</param>
		public NodeItemBase GetNewNode(PointF startPoint, ENodeRequirement nodeRequirement)
		{
			NodeItemBase nodeItem;

			// Instantiate the floorplan item selected via the tool strip
			if (tsbExternalWall.Checked)
				nodeItem = new NodeWallExternal(startPoint);
			else if (tsbInternalWall.Checked)
				nodeItem = new NodeWallInternal(startPoint);
			else if (tsbDoor.Checked)
				nodeItem = new NodeDoor();
			else if (tsbDesk.Checked)
				nodeItem = new NodeDesk(startPoint);
			else if (tsbTable.Checked)
				nodeItem = new NodeTable(startPoint);
			else if (tsbTextBox.Checked)
				nodeItem = new NodeTextBox(startPoint);
			else
			{
				ShowMessageBoxError("Select a floorplan item on the toolbar.");
				return null;
			}

			// Check the item specified by the tool strip buttons is suitable
			if (nodeRequirement == ENodeRequirement.eDrawable && !nodeItem.IsDrawable)
			{
				ShowMessageBoxError(
						"Floorplan item selected on the toolbar can not be drawn, it must be dropped.");
				return null;
			}
			if (nodeRequirement == ENodeRequirement.eDroppable && !nodeItem.IsDroppable)
			{
				ShowMessageBoxError(
						"Floorplan item selected on the toolbar can not be dropped, it must be drawn.");
				return null;
			}

			return nodeItem;
		}


		/// <summary>
		/// Method to display an error message in a MessageBox.
		/// </summary>
		public static void ShowMessageBoxError(string errorText)
		{
			MessageBox.Show(errorText, CProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
		}

		#endregion Public methods


		#region Private methods

		/// <summary>
		/// Method called when user selects menu item File - New (or hits Ctrl+N).
		/// 
		/// If an initial floorplan size is specified then four external walls are drawn. The user is 
		///  able to modify them if he/she changes his/her mind.
		/// </summary>
		private void File_New_Click(object sender, EventArgs e)
		{
			// Maybe save current floorplan, then create a completely new FloorplanDiagram object
			FloorplanDiagram floorplanDiagram = CreateBrandNewFloorplan();
			if (floorplanDiagram == null)
				return;

			// Display the modal settings form, let user change the settings
			(new FormSettings(floorplanDiagram, true)).ShowDialog();
			ApplySettings();

			// Draw external walls if optional initial size specified
			int sizeX = floorplanDiagram.RealWorldSizeX;
			int sizeY = floorplanDiagram.RealWorldSizeY;
			if (sizeX > 0 && sizeY > 0)
			{
				sizeX *= floorplanDiagram.ScalingFactor;
				sizeY *= floorplanDiagram.ScalingFactor;

				floorplanDiagram.Nodes.Add(new NodeWallExternal(new PointF(0f, 0f), sizeX, false));
				floorplanDiagram.Nodes.Add(new NodeWallExternal(new PointF(0f, 0f), sizeY, true));
				floorplanDiagram.Nodes.Add(new NodeWallExternal(new PointF(0f + sizeX, 0f), sizeY, true));
				floorplanDiagram.Nodes.Add(new NodeWallExternal(new PointF(0f, 0f + sizeY), sizeX, false));
			}

			// Create 20 pixel "pseudo margins" - see comments on the EstablishPseudoMargins() method
			EstablishPseudoMargins();
		}


		/// <summary>
		/// Method called when user selects menu item File - Open (or hits Ctrl+O).
		/// </summary>
		private void File_Open_Click(object sender, EventArgs e)
		{
			// Display an open file dialog suitable for opening a floorplan file
			using (OpenFileDialog openFileDialog = new OpenFileDialog())
			{
				openFileDialog.Title = "Open Floorplan File";
				openFileDialog.InitialDirectory =
										   Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
				openFileDialog.Filter =
				   "MindFusion Floorplan (*" + CFileNameExtension + ")|*" + CFileNameExtension + "|" +
				   "All Files (*.*)|*.*";
				DialogResult dialogResult = openFileDialog.ShowDialog();
				if (dialogResult == DialogResult.OK)
					SaveFloorplanFilename(openFileDialog.FileName);
				else
					return;
			}

			LoadFloorplan(_floorplanFilename);
		}

		void LoadFloorplan(string fileName)
		{
			// Maybe save current floorplan, then create a completely new FloorplanDiagram object
			FloorplanDiagram floorplanDiagram = CreateBrandNewFloorplan();
			if (floorplanDiagram == null)
				return;

			// Use MindFusion's file I/O to load the diagram from an XML file. This will invoke some 
			//  override methods in FloorplanDiagram that will extract the extra fields. It will also 
			//  invoke some override methods in some of the floorplan item classes to load their extra 
			//  fields.
			diagramView.Diagram.LoadFromXml(fileName);
			ApplySettings();

			// Create 20 pixel "pseudo margins" - see comments on the EstablishPseudoMargins() method
			EstablishPseudoMargins();
		}


		/// <summary>
		/// Method called when user selects menu item File - Save (or hits Ctrl+S).
		/// </summary>
		private void File_Save_Click(object sender, EventArgs e)
		{
			SaveTheFile(false);
		}


		/// <summary>
		/// Method called when user selects menu item File - Save As.
		/// </summary>
		private void File_SaveAs_Click(object sender, EventArgs e)
		{
			SaveTheFile(true);
		}


		/// <summary>
		/// Method called when user selects menu item File - Exit (or hits Alt+F4). The form is closed
		/// , and this results in a call to FormMain_FormClosing(), which may cancel the close.
		/// </summary>
		private void File_Exit_Click(object sender, EventArgs e)
		{
			Close();
		}


		/// <summary>
		/// Method called when user selectes menu item Edit - Select All (or hits Ctrl+A).
		/// </summary>
		private void Edit_SelectAll_Click(object sender, EventArgs e)
		{
			foreach (DiagramNode diagramNode in diagramView.Diagram.Nodes)
				diagramView.Diagram.Selection.AddItem(diagramNode);
		}


		/// <summary>
		/// Method called when user selectes menu item Edit - Select None. (The Esc key also does 
		/// this, but it is implemented separately because Esc can't be specified as a shortcut key.)
		/// </summary>
		private void Edit_SelectNone_Click(object sender, EventArgs e)
		{
			diagramView.Diagram.Selection.Clear();
		}


		/// <summary>
		/// Method called when user selectes menu item Edit - Delete (or hits Del key).
		/// </summary>
		private void Edit_Delete_Click(object sender, EventArgs e)
		{
			while (diagramView.Diagram.Selection.Items.Count > 0)
				diagramView.Diagram.Items.Remove(diagramView.Diagram.Selection.Items[0]);
		}


		/// <summary>
		/// Method called when user selectes menu item Edit - Copy (or hits Ctrl+C).
		/// </summary>
		private void Edit_Copy_Click(object sender, EventArgs e)
		{
			diagramView.CopyToClipboard(true);
			_pasteOffset = CPasteOffset;
		}


		/// <summary>
		/// Method called when user selectes menu item Edit - Cut (or hits Ctrl+X).
		/// </summary>
		private void Edit_Cut_Click(object sender, EventArgs e)
		{
			diagramView.CutToClipboard(true);
			_pasteOffset = 0;
		}


		/// <summary>
		/// Method called when user selectes menu item Edit - Paste (or hits Ctrl+V).
		/// </summary>
		private void Edit_Paste_Click(object sender, EventArgs e)
		{
			diagramView.PasteFromClipboard(_pasteOffset, _pasteOffset);
			_pasteOffset += CPasteOffset;
		}


		/// <summary>
		/// Method called when user selectes menu item View - Zoom In, or clicks the corresponding 
		/// toolbar button.
		/// </summary>
		private void View_ZoomIn_Click(object sender, EventArgs e)
		{
			diagramView.ZoomIn();
		}


		/// <summary>
		/// Method called when user selectes menu item View - Zoom Out, or clicks the corresponding 
		/// toolbar button.
		/// </summary>
		private void View_ZoomOut_Click(object sender, EventArgs e)
		{
			diagramView.ZoomOut();
		}


		/// <summary>
		/// Method called when user selectes menu item View - No Zoom, or clicks the corresponding 
		/// toolbar button.
		/// </summary>
		private void View_ZoomNone_Click(object sender, EventArgs e)
		{
			diagramView.ZoomFactor = 100f;
		}


		/// <summary>
		/// Method called when user selectes menu item View - Floorplan Settings.
		/// </summary>
		private void View_FloorplanSettings_Click(object sender, EventArgs e)
		{
			// Display the modal settings form, let user change the settings
			(new FormSettings(diagramView.Diagram as FloorplanDiagram, false)).ShowDialog();
			ApplySettings();
		}


		/// <summary>
		/// Method called when form is being closed, due to File - Exit or click on X in upper right 
		/// corner of the form.
		/// </summary>
		private void FormMain_FormClosing(object sender, FormClosingEventArgs e)
		{
			e.Cancel = !PromptForSaveIfNecessary();
		}


		/// <summary>
		/// Method to save the current FloorplanDiagram object if necessary / desired, and then to 
		/// create a completely new one. This is done to ensure there are no unexpected artifacts 
		/// lingering from the previous FloorplanDiagram object when providing the user with a new 
		/// empty floorplan.
		/// </summary>
		/// <returns>new FloorplanDiagram object, or null if user indicated cancel</returns>
		private FloorplanDiagram CreateBrandNewFloorplan()
		{
			if (!PromptForSaveIfNecessary())
				return null;

			diagramView.Diagram.Dispose();
			FloorplanDiagram floorplanDiagram = new FloorplanDiagram(this);
			diagramView.Diagram = floorplanDiagram;
			SaveFloorplanFilename(null);
			return floorplanDiagram;
		}


		/// <summary>
		/// Method to test if the current floorplan diagram (if any) is "dirty", and to prompt the 
		/// user asking if he/she wants to save it if so.
		/// </summary>
		/// <returns>true = all OK, false = user indicated cancel</returns>
		private bool PromptForSaveIfNecessary()
		{
			// Test if the floorplan diagram is "dirty" (has been modified since last saved)
			if (!diagramView.Diagram.Dirty)
				return true;

			// Prompt user to ask if he/she wants to save it or not
			DialogResult dialogResult = MessageBox.Show("Do you want to save the current floorplan?",
								 CProgramName, MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question);
			if (dialogResult == DialogResult.Cancel)
				return false;
			if (dialogResult == DialogResult.No)
				return true;

			// Save the floorplan diagram as an XML file
			SaveTheFile(false);
			return true;
		}


		/// <summary>
		/// Method to save the floorplan as an XML file, after perhaps prompting the user for a 
		/// filename and maybe a location.
		/// </summary>
		private void SaveTheFile(bool alwaysPrompt)
		{
			// Test if necessary to prompt user for a filename (and maybe a location)
			if (alwaysPrompt || _floorplanFilename == null)
			{
				// Prompt user to enter a file name, which will also provide a path
				using (SaveFileDialog saveFileDialog = new SaveFileDialog())
				{
					saveFileDialog.Title = "Save Floorplan File";
					saveFileDialog.InitialDirectory =
											Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
					if (_floorplanFilename != null)
					{
						saveFileDialog.FileName = SystemPath.GetFileName(_floorplanFilename);
						saveFileDialog.InitialDirectory = SystemPath.GetFullPath(_floorplanFilename);
					}
					saveFileDialog.Filter =
					   "MindFusion Floorplan (*" + CFileNameExtension + ")|*" + CFileNameExtension + "|" +
					   "All Files (*.*)|*.*";
					if (saveFileDialog.ShowDialog(this) != DialogResult.OK)
						return;
					SaveFloorplanFilename(saveFileDialog.FileName);
				}
			}

			// Use MindFusion's file I/O to save the diagram as an XML file. This will invoke some 
			//  override methods in FloorplanDiagram that will copy the extra fields to XML elements. 
			//  It will also invoke some override methods in some of the floorplan item classes to 
			//  persist their extra fields.
			diagramView.Diagram.SaveToXml(_floorplanFilename);
		}


		/// <summary>
		/// Method to save the floorplan name, and to adjust the form's title bar.
		/// </summary>
		/// <param name="floorplanFilename">floorplan file name, or null if none</param>
		private void SaveFloorplanFilename(string floorplanFilename)
		{
			_floorplanFilename = floorplanFilename;

			string s = CProgramName;
			if (floorplanFilename != null)
				s = SystemPath.GetFileNameWithoutExtension(floorplanFilename) + " - " + s;

			this.Text = s;
		}


		/// <summary>
		/// Method to apply the various settings in the "extra fields" of the FloorplanDiagram object 
		/// to the diagram itself and the DiagramView.
		/// </summary>
		private void ApplySettings()
		{
			FloorplanDiagram floorplanDiagram = diagramView.Diagram as FloorplanDiagram;
			if (floorplanDiagram == null)
				return;  // Should be impossible

			// Set up the ruler to show real-world dimensions in feet or meters
			if (floorplanDiagram.RealWorldUnits == ERealWorldUnits.eFeet)
			{
				mindFusionRuler.Text = "ft.";
				mindFusionRuler.Unit = new MeasureUnit("Foot", MeasureUnit.Pixel,
									floorplanDiagram.ScalingFactor, floorplanDiagram.ScalingFactor / 2);
			}
			else
			{
				mindFusionRuler.Text = "m.";
				mindFusionRuler.Unit = new MeasureUnit("Meter", MeasureUnit.Pixel,
													   floorplanDiagram.ScalingFactor);
			}

			// Test for a background image file specified and open it and apply it to the Diagram object
			if (floorplanDiagram.BackgroundImageFile != "" && floorplanDiagram.BackgroundOpacity != 0)
				ApplyBackgroundImage(floorplanDiagram);
			else
				floorplanDiagram.BackgroundImage = null;

			diagramView.Refresh();
		}


		/// <summary>
		/// Method to read the background image file and scale it and change its opacity and apply it 
		/// to the Diagram object.
		/// </summary>
		private static void ApplyBackgroundImage(FloorplanDiagram floorplanDiagram)
		{
			try
			{
				if (!File.Exists(floorplanDiagram.BackgroundImageFile))
					throw new FileNotFoundException(
										 "File " + floorplanDiagram.BackgroundImageFile + " not found.");

				Image backgroundImage = new Bitmap(floorplanDiagram.BackgroundImageFile);

				if (floorplanDiagram.BackgroundScaling != 100 ||
					floorplanDiagram.BackgroundOpacity != 100)
					backgroundImage = ImageUtilities.ResizeImage(backgroundImage,
															  floorplanDiagram.BackgroundScaling / 100f,
															  floorplanDiagram.BackgroundOpacity / 100f);

				floorplanDiagram.BackgroundImage = backgroundImage;
			}
			catch (Exception e)
			{
				MessageBox.Show("Error applying background image: " + e.Message,
								CProgramName, MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
		}


		/// <summary>
		/// Method called when user clicks on one of the ToolStrip buttons that specifies this event 
		/// handler as the Click event handler.
		/// </summary>
		private void ToolStripClickOnButton_Click(object sender, EventArgs e)
		{
			// Set the checked/unchecked status of all of the ToolStrip buttons
			foreach (ToolStripItem toolStripItem in toolStrip1.Items)
			{
				ToolStripButton toolStripButton = toolStripItem as ToolStripButton;
				if (toolStripButton != null)
					toolStripButton.Checked = toolStripButton == sender;
			}
		}


		/// <summary>
		/// Method called when the user (or the program) changes the scrolling. This detects the 
		/// situation when the user has scrolled down and then back up, or right and then back to the 
		/// left, and if so restablishes the 20 pixel "pseudo margins".
		/// 
		/// See comments on the EstablishPseudoMargins() method.
		/// </summary>
		private void DiagramView_ScrollChanged(object sender, EventArgs e)
		{
			if (diagramView.ScrollX == 0)
				diagramView.ScrollX = -CPseudoMarginSize;
			if (diagramView.ScrollY == 0)
				diagramView.ScrollY = -CPseudoMarginSize;
		}


		/// <summary>
		/// This method moves the rendering of the Diagram object down 20 pixels and left 20 pixels. 
		/// The intention is to establish a kind of "pseudo margins" of 20 pixels between the ruler 
		/// and the active drawing area in the DiagramView (and Diagram). This is desirable because 
		/// otherwise it is possible (or even probable) that an external wall item will be located 
		/// along the top and left edges of the drawing area, and when these are up against the ruler 
		/// it looks ugly, and also the wall items are partly hidden because their center line is at 
		/// 0,0, not their top or left edge. 
		/// 
		/// But this is a rather kludgey solution. It would be better if there was a way to specify a 
		/// margin or padding to seperate the ruler and the DiagramView, or to specify that the 0,0 
		/// point in the Diagram should be kept away from the upper left corner of the DiagramView.
		/// </summary>
		private void EstablishPseudoMargins()
		{
			diagramView.ScrollTo(-CPseudoMarginSize, -CPseudoMarginSize);
		}


		/// <summary>
		/// Method called when floorplan diagram area has focus and user hits a key on the keyboard. 
		/// This is used to implement de-selection of nodes when the Esc key is pressed. (This 
		/// function is also available from the menu: Edit - Select None.)
		/// </summary>
		private void DiagramView_KeyDown(object sender, KeyEventArgs e)
		{
			if (e.KeyCode == Keys.Escape)
			{
				diagramView.Diagram.Selection.Clear();
				e.SuppressKeyPress = true;  // Prevent child controls from handling this event as well 
			}
		}


		/// <summary>
		/// Method called by MindFusion to indicate inplace editing is done. 
		/// </summary>
		private void DiagramView_LeaveInplaceEditMode(object sender, InPlaceEditEventArgs e)
		{
			_inplaceEditingEnded = Environment.TickCount;
		}

		#endregion Private methods

		private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
		{
			new AboutForm().ShowDialog();
		}

	}
}